/*
** Module   : Registry.java
** Abstract : Progress 4GL compatible Winodws registry related methods
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 EVL 20130604 Created initial version.
*/

package com.goldencode.p2j.util;

import com.goldencode.util.*;

/**
 * A class implementing the Windows registry based environment methods. Key values are stored in
 * registry under appropriate keys in registry subsection. This is the default approach if none
 * new environment is not loaded and used by LOAD and USE Progress statements.
 *
 * @author    EVL
 */
public class Registry
implements EnvironmentAccessor
{
   /** Windows registry key-subkey names separator. */
   public static final String NAME_SEPARATOR = "\\";
   
   /** Windows defailt hive name. */
   public static final String HKEY_DEFAULT_BASE = "HKEY_CURRENT_USER";
   
   /** Windows defailt key name. */
   public static final String KEY_DEFAULT = "";
   
   /** Native entry point to get key value from registry. */
   private native static String getKeyVal(String base, String section, String key);
   
   /** Native entry point to store key value to registry. */
   private native static boolean setKeyVal(String base, String section, String key, String value);

   /** The complete key path with subkeys to store the key values without base key name. */
   private String registryKeyName = null;
   
   /** The name of the registry base to store the keys. */
   private String baseKeyName = null;

   /** The flag indicating whether we need to create new registry sections replacing existed. */   
   private boolean isNew;

   /** Load native library to handle registry related JNI calls on demand. */
   static
   {
      if (PlatformHelper.isUnderWindowsFamily())
      {
         System.loadLibrary("p2j");
      }
   }

   /**
    * Default constructor creating the internals based on the expected registry keys for currently
    * installed Progress simulation.
    */
   public Registry()
   {
      registryKeyName = "SOFTWARE"+NAME_SEPARATOR+"PSC"+NAME_SEPARATOR+"PROGRESS"+NAME_SEPARATOR+
                        EnvironmentOps.getVersion().toStringMessage();
      isNew = false;
      baseKeyName = HKEY_DEFAULT_BASE;
   }

   /**
    * Construct an instance of the environment with Windows registry as backend.
    *
    * @param env 
    *        The environment to create or load. Can be the registry key.
    * @param flagNew
    *        If <code>true</code> - the new file will be created overwriting the possibly
    *        existed one.
    * @param baseKey
    *        When the parameter is specified the serching is performed under the particular base
    *        key value.
    */
   public Registry(character env, logical flagNew, character baseKey)
   {
      registryKeyName = env.toStringMessage();
      isNew = flagNew.booleanValue(); 
      if (baseKey != null)
      {
         baseKeyName = baseKey.toStringMessage();
      }
      else
      {
         baseKeyName = HKEY_DEFAULT_BASE;
      }
   }

   /**
    * Creates application defaults implementing Progress LOAD statement.
    * 
    * @param env 
    *        The environment to create or load. Can be the registry key.
    * @param directory 
    *        The directory path for existing initialization file or place to create
    *        initialization file. Unused for registry related implementation.
    * @param flagNew
    *        If <code>true</code> - the new registry key will be created overwriting the possibly
    *        existed one.
    * @param baseKey
    *        When the parameter is specified the serching is performed under the particular base
    *        key value.
    */
   public void load(character env, character directory, logical flagNew, character baseKey)
   {
      // TODO implement this
      return;
   }

   /**
    * Specifies environment defaults to use in subsequent windows.
    * 
    * @param env 
    *        The environment to load. If the parameter is null string("") the default environment
    *        becomes the current one.
    */
   public void use(character env)
   {
      // TODO implement this
      return;
   }

   /**
    * Adds, modifies and deletes keys in current environment.
    * 
    * @param section 
    *        The name of the section containing the key to modify.
    * @param key 
    *        The name of the key key to modify or default key if not specified.
    * @param value 
    *        The new value of the key under modification.
    */
   public void setKeyValue(character section, character key, character value)
   {
      // TODO implement this
      // Prepare initial parameters for JNI call
      String sectionInitial = null;
      String keyInitial = null;
      String valueInitial = null;
      
      if (section != null)
      {
         sectionInitial = registryKeyName + NAME_SEPARATOR + section.toStringMessage();
      }
      else
      {
         sectionInitial = registryKeyName;
      }
      if (key != null)
      {
         keyInitial = key.toStringMessage();
      }
      else
      {
         keyInitial = KEY_DEFAULT;
      }
      if (value != null)
      {
         valueInitial = value.toStringMessage();
      }
      // Call C native code here
      setKeyVal(baseKeyName, sectionInitial, keyInitial, valueInitial);
   }

   /**
    * Getting the key value from the current environment.
    * 
    * @param section 
    *        The name of the section containing the key to get.
    * @param key 
    *        The name of the key key to get or default key if not specified.
    * 
    * @return The value of the key.
    */
   public character getKeyValue(character section, character key)
   {
      // TODO implement this
      // Prepare initial parameters for JNI call
      String sectionInitial = null;
      String keyInitial = null;
      String valueInitial = null;
      
      if (section != null)
      {
         sectionInitial = registryKeyName + NAME_SEPARATOR + section.toStringMessage();
      }
      else
      {
         sectionInitial = registryKeyName;
      }
      if (key != null)
      {
         keyInitial = key.toStringMessage();
      }
      else
      {
         keyInitial = KEY_DEFAULT;
      }
      // Call C native code here
      String result = getKeyVal(baseKeyName, sectionInitial, keyInitial);
      
      return new character(result);
   }

   /**
    * Unloads specifies environment from the current one.
    * 
    * @param env 
    *        The environment to unload. Can be the registry or initialization
    *        file. The environment to unload should be previously loaded by
    *        load() method.
    */
   public void unload(character env)
   {
      // TODO implement this
      return;
   }
}
