/*
** Module   : StanzaIni.java
** Abstract : Progress 4GL compatible Winodws INI files related methods
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 EVL 20130604 Created initial version.
*/

package com.goldencode.p2j.util;

import java.io.*;
import java.util.*;

/**
 * A class implementing the INI file based environment methods. The INI file is the plain text
 * file containing the inside the [] braces and key=value pairs within every section.
 *
 * @author    EVL
 */
public class StanzaIni
implements EnvironmentAccessor
{
   /** Extra urecognized data in INI file. This should be string never used for real key. */
   public static final String KEY_INRECOGNIZED_MARK = "----";
   
   /** The name of the INI file that holds the environment variables. */
   private String name = null;
   
   /** Working directory value to search the INI file. */
   private String workingDirectory = null;
   
   /** The flag indicating whether we need to create new file instead of opening existed one. */
   private boolean isNew = false;

   /** The map that caches the INI file contents. */
   private Map<String, Map<String, String>> envMap = null;
   
   /**
    * Default constructor creating the internal cache for environment.
    */
   public StanzaIni()
   {
      envMap = new LinkedHashMap<String, Map<String, String>>();
   }
   
   /**
    * Construct an instance of the environment with INI file as backend.
    *
    * @param env 
    *        The environment to create or load. The name without extension of the initialization
    *        file.
    * @param directory 
    *        The directory path for existing initialization file or place to create
    *        initialization file.
    * @param flagNew
    *        If <code>true</code> - the new file will be created overwriting the possibly
    *        existed one.
    */
   public StanzaIni(character env, character directory, logical flagNew)
   {
      this();

      name = env.toStringMessage() + ".ini";
      if (directory != null)
      {
         workingDirectory = directory.toStringMessage();
      }
      isNew = flagNew.booleanValue();
   }

   /**
    * Creates application defaults implementing Progress LOAD statement.
    * 
    * @param env 
    *        The environment to create or load. Initialization file name without extension.
    * @param directory 
    *        The directory path for existing initialization file or place to create
    *        initialization file.
    * @param flagNew
    *        If <code>true</code> - the new file will be created overwriting the possibly
    *        existed one.
    * @param baseKey
    *        When the parameter is specified the searching is performed under the particular key
    *        value. It is possible to specify "INI" as parameter. In this case search or creation
    *        will be dome inside initialization file specified in env parameter.
    */
   public void load(character env, character directory, logical flagNew, character baseKey)
   {
      // TODO implement this
      if (!isNew)
      {
         loadIniFile();
      }
      else
      {
         createEmptyIniFile();
      }
   }

   /**
    * Specifies environment defaults to use in subsequent windows.
    * 
    * @param env 
    *        The environment to load. If the parameter is null string("") the
    *        default environment becomes the current one.
    */
   public void use(character env)
   {
      // TODO implement this
      return;
   }

   /**
    * Adds, modifies and deletes keys in current environment.
    * 
    * @param section 
    *        The name of the section containing the key to modify.
    * @param key 
    *        The name of the key key to modify or default key if not specified.
    * @param value 
    *        The new value of the key under modification.
    */
   public void setKeyValue(character section, character key, character value)
   {
      // TODO implement this
      String sectionAsString = section.toStringMessage().trim();
      String keyAsString = key.toStringMessage().trim();
      // The 4GL does allow to put extra spaces for value string into INI file
      String valueAsString = value.toStringMessage();
      
      Map<String, String> sectionBody = envMap.get(sectionAsString);
      if (sectionBody == null)
      {
         sectionBody = new LinkedHashMap<String, String>();
      }
      sectionBody.put(keyAsString, valueAsString);
      envMap.put(sectionAsString, sectionBody);
      
      commitIniFile();
   }

   /**
    * Getting the key value from the current environment.
    * 
    * @param section 
    *        The name of the section containing the key to get.
    * @param key 
    *        The name of the key key to get or default key if not specified.
    * 
    * @return The value of the key.
    */
   public character getKeyValue(character section, character key)
   {
      // TODO implement this
      String sectionAsString = section.toStringMessage().trim();
      String keyAsString = key.toStringMessage().trim();
      String valueAsString = null;
      
      Map<String, String> sectionBody = envMap.get(sectionAsString);
      if (sectionBody != null)
      {
         valueAsString = sectionBody.get(keyAsString);
      }
      // For missing entries 4GL returns empty string
      if (valueAsString == null)
      {
         return new character("");
      }
      else
      {
         // For statement GET-KEY-VALUE 4GL does not allow to return extra spaces unlike to
         // PUT-KEY-VALUE. Pretty strange behavior but was tested on real Windows 4GL system.
         return new character(valueAsString.trim());
      }
   }

   /**
    * Unloads specifies environment from the current one.
    * 
    * @param env 
    *        The environment to unload. For INI file it is initialization file name without
    *        extension. The environment to unload should be previously loaded by load() method.
    */
   public void unload(character env)
   {
      // TODO implement this
      return;
   }

   /**
    * Reconstructs full initial file name with possible path.
    */
   private String getFullIniFileName()
   {
      StringBuilder fullName = new StringBuilder();
      
      // We have non-empty directory value, add it first 
      if (workingDirectory != null && workingDirectory.length() > 0)
      {
         fullName.append(workingDirectory);
         fullName.append(EnvironmentOps.getLegacyFileSeparator());
      }
      // And we have the non-empty file name, add it
      if (name != null && name.length() > 0)
      {
         fullName.append(name);
      }
      // If something was added, convert to String
      if (fullName.length() > 0)
      {
         return fullName.toString();
      }
      else // Otherwise return null
      {
         return null;
      }
   }
   
   /**
    * Reads the current environment set from appropriate INI file in file system.
    */
   private void loadIniFile()
   {
      String fileName = getFullIniFileName();
      
      if (fileName != null && fileName.length() > 0)
      {
         BufferedReader br = null;
         
         try
         {
            // Open file stream for reading
            br = new BufferedReader(new FileReader(fileName));

            String sectionName = null;
            Map<String, String> sectionBody = null;
            
            String nextLine = br.readLine();
            // Read until EOF(nextLine == null)
            while (nextLine != null)
            {
               // Next INI file section should start here
               if (nextLine.length() > 2 && nextLine.charAt(0) == '[' &&
                   nextLine.charAt(nextLine.length() - 1) == ']')
               {
                  // The previously constructed section should be saved before starting to
                  // process the next section
                  if (sectionName != null && sectionBody != null && sectionBody.size() > 0)
                  {
                     envMap.put(sectionName, sectionBody);
                  }
                  // The new section name candidate
                  sectionName = new String(nextLine.substring(1,nextLine.length() - 1));
                  // Purge the previous one if it was existed
                  if (sectionBody != null)
                  {
                     sectionBody = null;
                  }
               }
               // Section processing
               else
               {
                  // Key-value candidate, get the index of separator first
                  int separatorNdx = nextLine.indexOf('=');
                  // Something should exist before and after separator
                  if (separatorNdx > 0 && nextLine.length() > separatorNdx + 1)
                  {
                     if (sectionBody == null)
                     {
                        sectionBody = new LinkedHashMap<String, String>();
                     }
                     // Construct key-value pair from line string
                     // Key name should be trimmed, need to check other cases
                     String keyName = new String(nextLine.substring(0,separatorNdx).trim());
                     // The value should remain as it is in INI file on reading stage.
                     String keyValue = new String(nextLine.substring(separatorNdx + 1));
                     // Store the next  key-value pair into section map
                     sectionBody.put(keyName, keyValue);
                  }
                  // Here we add artificial meaningless data including empty lines between
                  // sections
                  else
                  {
                     long i = 0; // Key modifier
                     String spaceLineKey = new String(KEY_INRECOGNIZED_MARK+i);
                     if (sectionBody == null)
                     {
                        sectionBody = new LinkedHashMap<String, String>();
                     }
                     // Create new unique key name for new line to be added
                     while (sectionBody.containsKey(spaceLineKey))
                     {
                        spaceLineKey = new String(KEY_INRECOGNIZED_MARK+(++i));
                     }
                     // Put "empty" lines into section map as if it is usual key
                     sectionBody.put(spaceLineKey, nextLine);
                  }
               }
               
               // Read the next line from file
               nextLine = br.readLine();
               if (nextLine == null && sectionName != null &&
                   sectionBody != null && sectionBody.size() > 0)
               {
                  // Possible last section in the file
                  envMap.put(sectionName, sectionBody);
               }
            }
            br.close();
         }
         catch (IOException ioex)
         {
         }
      }
   }
   
   /**
    * Creates the empty INI file for current environment set in file system.
    */
   private void createEmptyIniFile()
   {
      String fileName = getFullIniFileName();
      
      if (fileName != null && fileName.length() > 0)
      {
         PrintWriter pw = null;
         
         try
         {
            // Open file for writing
            pw = new PrintWriter(new FileWriter(fileName, false));
         }
         catch (IOException ioex)
         {
         }
         finally
         {
            // Do nothing and close it
            if (pw != null)
            {
               pw.close();
            }
         }
      }
   }
   
   /**
    * Writes the current environment set to the appropriate INI file in file system.
    */
   private void commitIniFile()
   {
      String fileName = getFullIniFileName();
      
      if (fileName != null && fileName.length() > 0)
      {
         PrintWriter pw = null;
         
         try
         {
            // Open file for writing
            pw = new PrintWriter(new FileWriter(fileName, false));
            // Sections iterator
            Iterator iterSect = envMap.keySet().iterator();
            while (iterSect.hasNext())
            {
               // Get next available section name
               String sectionName = (String)iterSect.next();
               Map<String, String> sectionBody = envMap.get(sectionName);
               // If section name is not empty write the section name
               if (sectionBody.size() > 0)
               {
                  pw.println("[" + sectionName + "]");
               }
               // Key names iterator
               Iterator iterKey = sectionBody.keySet().iterator();
               while (iterKey.hasNext())
               {
                  // Get next available key=value pair
                  String keyName = (String)iterKey.next();
                  String keyValue = sectionBody.get(keyName);
                  // And write them to the file
                  // Unrecognized line, write it as it was in INI file
                  if (keyName.indexOf(KEY_INRECOGNIZED_MARK) != -1)
                  {
                     pw.println(keyValue);
                  }
                  // Usual key-value pair
                  else
                  {
                     pw.println(keyName + "=" + keyValue);
                  }
               }
            }
            // Commit the changes
            pw.flush();
         }
         catch (IOException ioex)
         {
         }
         finally
         {
            if (pw != null)
            {
               // And finally close the file
               pw.close();
            }
         }
      }
   }
}
